// -------------------------------------------------
// CC Nue analysis module
//
// Author: Rory Fitzpatrick
// Updated: 2/19/18
// -------------------------------------------------

#ifndef CCNUE_H
#define CCNUE_H

// Framework includes
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/View.h"
#include "canvas/Persistency/Common/Ptr.h"
#include "canvas/Persistency/Common/PtrVector.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "canvas/Persistency/Common/FindMany.h"

#include "T962_Objects/MINOS.h"
#include "T962_Objects/Paddles.h"

#include "larcore/Geometry/Geometry.h"
#include "nusimdata/SimulationBase/MCTruth.h"
#include "nusimdata/SimulationBase/MCFlux.h"
#include "lardataobj/Simulation/SimChannel.h"
#include "lardataobj/AnalysisBase/Calorimetry.h"
#include "lardataobj/AnalysisBase/ParticleID.h"
#include "lardataobj/RawData/RawDigit.h"
#include "lardataobj/RawData/BeamInfo.h"
#include "lardata/Utilities/AssociationUtil.h"
#include "lardata/DetectorInfoServices/DetectorPropertiesService.h"
#include "larcoreobj/SummaryData/POTSummary.h"
#include "larsim/MCCheater/BackTrackerService.h"
#include "larsim/MCCheater/ParticleInventoryService.h"
#include "lardataobj/RecoBase/Track.h"
#include "lardataobj/RecoBase/TrackHitMeta.h"
#include "lardataobj/RecoBase/Cluster.h"
#include "lardataobj/RecoBase/Shower.h"
#include "lardataobj/RecoBase/Wire.h"
#include "lardataobj/RecoBase/Hit.h"
#include "lardataobj/RecoBase/EndPoint2D.h"
#include "lardataobj/RecoBase/Vertex.h"
#include "larcoreobj/SimpleTypesAndConstants/geo_types.h"
#include "larreco/Calorimetry/CalorimetryAlg.h"

#include <vector>
#include <cmath> 
#include <string>
#include <sstream>
#include <fstream>
#include <algorithm>
#include <functional>

#include "TTree.h"
#include "TFile.h"
#include "TH1.h"
#include "TSystem.h"
#include "TString.h"
#include "TTimeStamp.h"

//const int kMaxTracks     = 1000;  //maximum number of tracks                                                   
//const int kMaxClusters   = 1000;  //maximum number of clusters
//const int kMaxShowers    = 1000;  //maximum number of showers                                                  
//const int kMaxHits       = 20000; //maximum number of hits
//const int kMaxPrimaries  = 20000;  //maximum number of primary particles
//const int kMaxTrackHits  = 1000;  //maximum number of hits on a track
namespace t962 {

  class CCNueSelection : public art::EDAnalyzer {

  public:

    explicit CCNueSelection(fhicl::ParameterSet const& pset);
    virtual ~CCNueSelection();

    void reconfigure(fhicl::ParameterSet const& pset);
    void beginJob();
    void analyze(const art::Event& evt);
    
  protected:

  private: 

    void resetVars();

    TTree* fTree;
    // run information
    int    run;
    int    subrun;
    int    event;
    double evttime;
    double beamtime;
    double pot;
    double taulife;
    int    isdata;

    //reconstructed information 
    double vtxx_reco;
    double vtxy_reco;
    double vtxz_reco;
    double enu_reco;

    int    ntracks_reco;         //number of reconstructed tracks    

    int    nclusu_reco;
    int    nclusv_reco;
    int    clusID;
    int    nshwcls;

    //mctruth information 
    int    nuPDG_truth;
    int    ccnc_truth;
    int    mode_truth;
    double enu_truth;
    double nuvtxx_truth;
    double nuvtxy_truth;
    double nuvtxz_truth;

    // parameter set
    std::string fDigitModuleLabel;
    std::string fHitsModuleLabel;
    std::string fLArG4ModuleLabel;
    std::string fCalDataModuleLabel;
    std::string fGenieGenModuleLabel;
    std::string fG4ModuleLabel;
    std::string fClusterModuleLabel;
    std::string fShowerModuleLabel;
    std::string fTrackModuleLabel;
    std::string fEndPoint2DModuleLabel;
    std::string fVertexModuleLabel;
    std::string fMINOSModuleLabel;
    std::string fTrackMatchModuleLabel;
    std::string fPOTModuleLabel;
    std::string fCalorimetryModuleLabel;
    std::string fParticleIDModuleLabel;
    std::string fPaddlesModuleLabel;

    double fvertextrackWindow;
    double fvertexclusterWindow;
    calo::CalorimetryAlg fCalorimetryAlg;

  }; // class CCNueSelection

} // t962

// -------------------------------------------------

t962::CCNueSelection::CCNueSelection(fhicl::ParameterSet const& pset) :
  EDAnalyzer(pset), 
  fCalorimetryAlg(pset.get<fhicl::ParameterSet>("CalorimetryAlg")) {
  this->reconfigure(pset);
} // CCNueSelection

// -------------------------------------------------

t962::CCNueSelection::~CCNueSelection() {
} // ~CCNueSelection

// -------------------------------------------------

void t962::CCNueSelection::reconfigure(fhicl::ParameterSet const& pset) {
  fDigitModuleLabel = pset.get< std::string >("DigitModuleLabel");
  fHitsModuleLabel = pset.get< std::string >("HitsModuleLabel");
  fLArG4ModuleLabel = pset.get< std::string >("LArGeantModuleLabel");
  fCalDataModuleLabel = pset.get< std::string >("CalDataModuleLabel");
  fGenieGenModuleLabel = pset.get< std::string >("GenieGenModuleLabel");
  fG4ModuleLabel = pset.get< std::string >("G4ModuleLabel");
  fClusterModuleLabel = pset.get< std::string >("ClusterModuleLabel");
  fShowerModuleLabel = pset.get< std::string >("ShowerModuleLabel");
  fTrackModuleLabel = pset.get< std::string >("TrackModuleLabel");
  fEndPoint2DModuleLabel = pset.get< std::string >("EndPoint2DModuleLabel");
  fVertexModuleLabel = pset.get< std::string >("VertexModuleLabel");
  fMINOSModuleLabel = pset.get< std::string >("MINOSModuleLabel");
  fTrackMatchModuleLabel = pset.get< std::string >("TrackMatchModuleLabel");
  fPOTModuleLabel = pset.get< std::string >("POTModuleLabel");
  fCalorimetryModuleLabel = pset.get< std::string >("CalorimetryModuleLabel");
  fParticleIDModuleLabel = pset.get< std::string >("ParticleIDModuleLabel");
  fPaddlesModuleLabel = pset.get< std::string >("PaddlesModuleLabel");
  fvertextrackWindow = pset.get< double >("vertextrackWindow");
  fvertexclusterWindow = pset.get< double >("vertexclusterWindow");
} // reconfigure

// -------------------------------------------------

void t962::CCNueSelection::beginJob() {
  art::ServiceHandle<art::TFileService> tfs;
  fTree = tfs->make<TTree>("nueselection", "cc nue selection");
  fTree->Branch("run", &run, "run/I");
  fTree->Branch("subrun",&subrun,"subrun/I");
  fTree->Branch("event",&event,"event/I");
  fTree->Branch("evttime",&evttime,"evttime/D");
  fTree->Branch("beamtime",&beamtime,"beamtime/D");
  fTree->Branch("pot",&pot,"pot/D");
  fTree->Branch("taulife",&taulife,"taulife/D");
  fTree->Branch("isdata",&isdata,"isdata/I");

  fTree->Branch("vtxx_reco",&vtxx_reco,"vtxx_reco/D");
  fTree->Branch("vtxy_reco",&vtxy_reco,"vtxy_reco/D");
  fTree->Branch("vtxz_reco",&vtxz_reco,"vtxz_reco/D");
  fTree->Branch("enu_reco",&enu_reco,"enu_reco/D");

  fTree->Branch("ntracks_reco",&ntracks_reco,"ntracks_reco/I");

  fTree->Branch("nclusu_reco",&nclusu_reco,"nclusu_reco/I");
  fTree->Branch("nclusv_reco",&nclusv_reco,"nclusv_reco/I");
  fTree->Branch("clusID",&clusID,"clusID/I");
  fTree->Branch("nshwcls",&nshwcls,"nshwcls/I");

  fTree->Branch("nuPDG_truth",&nuPDG_truth,"nuPDG_truth/I");
  fTree->Branch("ccnc_truth",&ccnc_truth,"ccnc_truth/I");
  fTree->Branch("mode_truth",&mode_truth,"mode_truth/I");
  fTree->Branch("enu_truth",&enu_truth,"enu_truth/D");
  fTree->Branch("nuvtxx_truth",&nuvtxx_truth,"nuvtxx_truth/D");
  fTree->Branch("nuvtxy_truth",&nuvtxy_truth,"nuvtxy_truth/D");
  fTree->Branch("nuvtxz_truth",&nuvtxz_truth,"nuvtxz_truth/D");

} // beginJob

// -------------------------------------------------

void t962::CCNueSelection::analyze(const art::Event& evt) {

  resetVars();
  
  run = evt.run();
  subrun = evt.subRun();
  event = evt.id().event();

  art::Timestamp ts = evt.time();
  TTimeStamp tts(ts.timeHigh(), ts.timeLow());

  art::Handle<raw::BeamInfo> beam;
  if (evt.getByLabel("beam", beam)) {
    beamtime = (double)beam->get_t_ms();
    beamtime /= 1000.; // in seconds
  }

  if (evt.isRealData()) isdata = 1;
  else isdata = 0;

  // hits, wires, tracks, clusters, showers, vetices, endpoints
  art::Handle< std::vector<recob::Hit> > hitListHandle;
  std::vector<art::Ptr<recob::Hit> > hitlist;
  if (evt.getByLabel(fHitsModuleLabel,hitListHandle))
    art::fill_ptr_vector(hitlist, hitListHandle);

  art::Handle< std::vector<recob::Track> > trackListHandle;
  std::vector<art::Ptr<recob::Track> > tracklist;
  if (evt.getByLabel(fTrackModuleLabel,trackListHandle))
    art::fill_ptr_vector(tracklist, trackListHandle);

  art::Handle< std::vector<recob::EndPoint2D> > endpointListHandle;
  std::vector<art::Ptr<recob::EndPoint2D> > endpointlist;
  if (evt.getByLabel(fEndPoint2DModuleLabel,endpointListHandle))
    art::fill_ptr_vector(endpointlist, endpointListHandle);
  
  art::Handle< std::vector<recob::Vertex> > vertexListHandle;
  std::vector<art::Ptr<recob::Vertex> > vertexlist;
  if (evt.getByLabel(fVertexModuleLabel,vertexListHandle))
    art::fill_ptr_vector(vertexlist, vertexListHandle);

  art::Handle< std::vector<recob::Cluster> > clusterListHandle;
  std::vector<art::Ptr<recob::Cluster> > clusterlist;
  if (evt.getByLabel(fClusterModuleLabel,clusterListHandle))
    art::fill_ptr_vector(clusterlist, clusterListHandle);

  art::Handle< std::vector<recob::Shower> > shwListHandle;
  std::vector<art::Ptr<recob::Shower> > shwlist;
  if (evt.getByLabel(fShowerModuleLabel,shwListHandle))
    art::fill_ptr_vector(shwlist, shwListHandle);

  // minos
  art::Handle< std::vector<t962::MINOS> > minosListHandle;
  std::vector<art::Ptr<t962::MINOS> > minoslist;
  if (evt.getByLabel(fMINOSModuleLabel,minosListHandle))
    art::fill_ptr_vector(minoslist, minosListHandle);
  
  // mc info
  art::Handle< std::vector<simb::MCTruth> > mctruthListHandle;
  std::vector<art::Ptr<simb::MCTruth> > mclist;
  if (evt.getByLabel(fGenieGenModuleLabel,mctruthListHandle))
    art::fill_ptr_vector(mclist, mctruthListHandle);

  // get truth information
  if (!isdata && mclist.size()) {
    art::Ptr<simb::MCTruth> mctruth = mclist[0];
    if (mctruth->NeutrinoSet()) {
      ccnc_truth = mctruth->GetNeutrino().CCNC();
      mode_truth = mctruth->GetNeutrino().Mode();
      nuPDG_truth = mctruth->GetNeutrino().Nu().PdgCode();
      enu_truth = mctruth->GetNeutrino().Nu().E();
      nuvtxx_truth = mctruth->GetNeutrino().Nu().Vx();
      nuvtxy_truth = mctruth->GetNeutrino().Nu().Vy();
      nuvtxz_truth = mctruth->GetNeutrino().Nu().Vz();
    }
  }

  // primitive vertex reco (seriously, it's really primitive)
  if (vertexlist.size()) {
    double vtxxyz[3];
    vertexlist[0]->XYZ(vtxxyz);
    vtxx_reco = vtxxyz[0];
    vtxy_reco = vtxxyz[1];
    vtxz_reco = vtxxyz[2];
  }

  ntracks_reco = tracklist.size();
  
  //  art::FindManyP<recob::Hit> trk_fm(trackListHandle, evt, fTrackModuleLabel);
  //  art::FindManyP<recob::Hit> cls_fm(clusterListHandle, evt, fClusterModuleLabel);

  //  auto const* detprop = lar::providerFrom<detinfo::DetectorPropertiesService>();
  //  art::ServiceHandle<geo::Geometry> geom;     
 
  for (size_t i = 0; i < tracklist.size(); ++i){
    //    std::vector< art:: Ptr<recob::Hit> > trk_hitlist = trk_fm.at(tracklist[i]);
    for (size_t j = 0; j < clusterlist.size(); ++j) {
      if (clusterlist[j]->View() == geo::kU) continue;
      //      std::vector< art::Ptr<recob::Hit> > cls_hitlist = cls_fm.at(clusterlist[i]);

      //      for (size_t ii = 0; ii < trk_hitlist.size(); ++ii) {
      //	if (trk_hitlist[ii]->View() == geo::kU) continue;
      //	for (size_t jj = 0; jj < cls_hitlist.size(); ++jj) {

	  // TEST THIS CODE!!
	  /*
	  float wirePitch = geom->WirePitch(geo::kV);
	  float tickToDist = detprop->DriftVelocity(detprop->Efield(),detprop->Temperature());
	  tickToDist *= 1.e-3 * detprop->SamplingRate(); // 1e-3 is conversion of 1/us to 1/ns                     
	  UnitsPerTick = tickToDist / WirePitch;

	  float wire1 = trk_hitlist[ii]->WireID().Wire;
	  float time1 = trk_hitlist[ii].PeakTime * UnitsPerTick;

	  float wire2 = cls_hitlist[jj]->WireID().Wire;
	  float time2 = cls_hitlist[jj].PeakTime * UnitsPerTick;

	  float dist = std::sqrt((wire1-wire2)*(wire1-wire2) + (time1-time2)*(time1-time2));
	  */
	  // TEST THIS CODE

      //	} // loop through cluster hitlist
      //      } // loop through track hitlist

    } // loop through clusters
  } // loop through tracks

  nclusu_reco = 0;
  nclusv_reco = 0;
  nshwcls = 0;
  // count clusters on each plane
  for (size_t i = 0; i < clusterlist.size(); ++i) {
    clusID = clusterlist[i]->ID();
    if (clusID < 0) nshwcls++;
    switch (clusterlist[i]->View()) {
    case geo::kU :
      nclusu_reco++; break;
    case geo::kV :
      nclusv_reco++; break;
    default : break;
    } // switch view
  }

  fTree->Fill();

} // analyze

// -------------------------------------------------

void t962::CCNueSelection::resetVars() {

  run = -99999;
  subrun = -99999;
  event = -99999;
  evttime = -99999;
  beamtime = -99999;
  isdata = -99999;
  taulife = -99999;

  vtxx_reco = -99999;
  vtxy_reco  = -99999;
  vtxz_reco  = -99999;
  
  ntracks_reco = -99999;

  nclusu_reco  = -99999;
  nclusv_reco  = -99999;
  clusID = -99999;
  nshwcls = -99999;

  nuPDG_truth = -99999;
  ccnc_truth = -99999;
  mode_truth = -99999;
  enu_truth = -99999;
  nuvtxx_truth = -99999;
  nuvtxy_truth = -99999;
  nuvtxz_truth = -99999;

} // resetVars

// -------------------------------------------------

namespace t962 {

  DEFINE_ART_MODULE(CCNueSelection)

} // t962

#endif
